unit u_math;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, u_myfloat;

type
  TObjTyp = (Obj_none,Obj_Value,Obj_Operator);

type
  TObj = ^ Objrec;
  Objrec = record
   Value : double;
   c     : char;
   typ   : TObjTyp;
   end;

type
  TFormel = class(TObject)
  private
    { Private-Deklarationen }
    FormelList : TList;
    Value : double;
    function calc(Liste : TList) : double;
    function CopyData(pData : TObj) : TObj;
    function getBracket(Liste : TList) : boolean;
    function getMultiplicator(Liste : TList) : boolean;
  public
    { Public-Deklarationen }
    Constructor Create;
    Destructor Destroy;
    procedure Clear;
    procedure Add(f : double); Overload;
    procedure Add(c : char); Overload;
    procedure Add(s : string); Overload;
    function getLastChar : char;
    function getLastValue : double;
    function getValue : double;
    function getString : string; Overload;
    function getString(List : TList) : string; Overload;
    function Count : integer;
    function getLastItemTyp : TObjTyp;
    function getLastItemIsValue : boolean;
    function getLastItemIsOperator : boolean;
    procedure Calculate;
  end;

implementation

Constructor TFormel.Create;
begin
  inherited Create;
  FormelList:= TList.Create;
  Clear;
end;

Destructor TFormel.Destroy;
begin
  Clear;
  FormelList.Free;
  inherited;
end;

procedure TFormel.Clear;
var i     : integer;
    pData : TObj;
begin
  Value:= 0.0;
  if FormelList.Count > 0 then
  begin
    for i:= 0 to FormelList.Count-1 do
    begin
      pData:= FormelList.Items[i];
      if pData <> nil then dispose(pData);
    end;
    FormelList.Clear;
  end;
end;

function TFormel.getLastChar : char;
var pData : TObj;
begin
  if FormelList.Count > 0 then
  begin
    pData:= FormelList.Items[FormelList.count-1];
    if pData <> nil then getLastChar:= pData.c
                    else getLastChar:= ' ';
  end
  else getLastChar:= ' ';
end;

function TFormel.getLastValue : double;
var pData : TObj;
begin
  if FormelList.Count > 0 then
  begin
    pData:= FormelList.Items[FormelList.count-1];
    if pData <> nil then getLastValue:= pData.Value
                    else getLastValue:= 0.0;
  end
  else getLastValue:= 0.0;
end;

function TFormel.getValue : double;
begin
  getValue:= Value;
end;

function TFormel.getString : string;
var i     : integer;
    pData : TObj;
    s     : string;
begin
  s:= '';
  if FormelList.Count > 0 then
  begin
    for i:= 0 to FormelList.Count-1 do
    begin
      pData:= FormelList.Items[i];
      if pData <> nil then
      begin
        case pData.typ of
        Obj_None:     begin
                      end;
        Obj_Value:    begin
                        s:= s + my_doubletostr(pData.Value,4,5);
                      end;
        Obj_Operator: begin
                        s:= s + pData.c;
                      end;
        end; //case

      end;
    end;
  end;
  getString:= s;
end;

function TFormel.getString(List : TList) : string;
var i     : integer;
    pData : TObj;
    s     : string;
begin
  s:= '';
  if List.Count > 0 then
  begin
    for i:= 0 to List.Count-1 do
    begin
      pData:= List.Items[i];
      if pData <> nil then
      begin
        case pData.typ of
        Obj_None:     begin
                      end;
        Obj_Value:    begin
                        s:= s + my_doubletostr(pData.Value,4,5);
                      end;
        Obj_Operator: begin
                        s:= s + pData.c;
                      end;
        end; //case

      end;
    end;
  end;
  getString:= s;
end;

procedure TFormel.Add(f : double);
var pData : TObj;
begin
  new(pData);
  pData.Value:= f;
  pData.c:= ' ';
  pData.typ:= Obj_Value;
  FormelList.Add(pData);
end;

procedure TFormel.Add(c : char);
var pData : TObj;
begin
  new(pData);
  pData.Value:= 0.0;
  pData.c:= c;
  pData.typ:= Obj_Operator;
  FormelList.Add(pData);
end;

procedure TFormel.Add(s : string);
var i     : integer;
    pData : TObj;
begin
  if s <> '' then
  begin
    for i:= 1 to length(s) do
    begin
      new(pData);
      pData.Value:= 0.0;
      pData.c:= s[i];
      pData.typ:= Obj_Operator;
      FormelList.Add(pData);
    end;
  end;
end;

function TFormel.Count : integer;
begin
  Count:= FormelList.Count;
end;

function TFormel.getLastItemTyp : TObjTyp;
var pData : TObj;
begin
  getLastItemTyp:= Obj_none;
  if FormelList.Count > 0 then
  begin
    pData:= FormelList.Items[FormelList.count-1];
    if pData <> nil then getLastItemTyp:= pData.typ;
  end;
end;

function TFormel.getLastItemIsValue : boolean;
var pData : TObj;
begin
  getLastItemIsValue:= false;
  if FormelList.Count > 0 then
  begin
    pData:= FormelList.Items[FormelList.count-1];
    if (pData <> nil) and (pData.typ = Obj_Value) then getLastItemIsValue:= true;
  end;
end;

function TFormel.getLastItemIsOperator : boolean;
var pData : TObj;
begin
  getLastItemIsOperator:= false;
  if FormelList.Count > 0 then
  begin
    pData:= FormelList.Items[FormelList.count-1];
    if (pData <> nil) and (pData.typ = Obj_Operator) then getLastItemIsOperator:= true;
  end;
end;

procedure TFormel.Calculate;
var pData : TObj;
begin
  if (FormelList <> nil) and (FormelList.Count > 0) then
  begin
    pData:= FormelList.Items[0];
    if (pData.c = '+') and (pData.typ = Obj_Operator) then
    begin
      dispose(pData);
      FormelList.Delete(0);
    end;
  end;

  Value:= calc(FormelList);
end;

function TFormel.CopyData(pData : TObj) : TObj;
var pNewData : TObj;
begin
  new(pNewData);
  if (pNewData <> nil) then
  begin
    if (pData <> nil) then
    begin
      pNewData.Value:= pData.Value;
      pNewData.c:= pData.c;
      pNewData.typ:= pData.typ;
    end
    else begin
           pNewData.Value:= 0.0;
           pNewData.c:= ' ';
           pNewData.typ:= Obj_none;
         end;
  end
  else pNewData:= nil;
  CopyData:= pNewData;
end;

function TFormel.getBracket(Liste : TList) : boolean;
var i     : integer;
    pData : TObj;
begin
  getBracket:= false;
  if (Liste <> nil) and (Liste.Count > 0) then
  begin
    i:= 0;
    while i < Liste.Count do
    begin
      pData:= Liste.Items[i];
      if pData <> nil then
      begin
        if (pData.typ = Obj_Operator) and (pData.c in ['(',')','[',']']) then
        begin
          getBracket:= true;
          break;
        end;
      end;
      inc(i);
    end;
  end;
end;

function TFormel.getMultiplicator(Liste : TList) : boolean;
var i     : integer;
    pData : TObj;
begin
  getMultiplicator:= false;
  if (Liste <> nil) and (Liste.Count > 0) then
  begin
    i:= 0;
    while i < Liste.Count do
    begin
      pData:= Liste.Items[i];
      if pData <> nil then
      begin
        if (pData.typ = Obj_Operator) and (pData.c = '*') then
        begin
          getMultiplicator:= true;
          break;
        end;
      end;
      inc(i);
    end;
  end;
end;

function TFormel.calc(Liste : TList) : double;
var i,n,k,x        : integer;
    pData,pNewData : TObj;
    newList        : TList;
    fValue         : double;
begin
  calc:= 0.0;
  if (Liste <> nil) and (Liste.count > 0) then
  begin
    while getBracket(Liste) do
    begin
      i:= 0;
      n:= 0;
      k:= 0;
      x:= 0;
      fValue:=0.0;
      //Klammern
      while i < Liste.Count do
      begin
        pData:= Liste.Items[i];
        if pData <> nil then
        begin
          case pData.typ of
          Obj_none:     begin
                        end;
          Obj_Value:    begin
                        end;
          Obj_Operator: begin
                          if pData.c = '(' then
                          begin
                            if k = 0 then x:= i;
                            inc(k);
                          end;

                          if pData.c = ')' then
                          begin
                            dec(k);
                            if k = 0 then
                            begin
                              newList:= TList.Create;
                              for n:= x+1 to i-1 do
                              begin
                                new(pNewData);
                                pData:= Liste.Items[n];
                                pNewData:= CopyData(pData);
                                newList.Add(pNewData);
                              end;
                              //iterativer Aufruf
                              fValue:= calc(newList);

                              for n:= i downto x do
                              begin
                                pData:= Liste.Items[n];
                                dispose(pData);
                                Liste.Delete(n);
                              end;

                              new(pNewData);
                              pNewData.Value:= fValue;
                              pNewData.c:= ' ';
                              pNewData.typ:= Obj_Value;
                              Liste.Insert(x,pNewData);

                              i:= -1; //restart
                            end;
                          end;
                        end;
          end; //case
        end;
        inc(i);
      end; //while
    end;

    //Punktrechnung
    while getMultiplicator(Liste) do
    begin
      i:= 0;
      n:= 0;
      k:= 0;
      x:= 0;
      fValue:=0.0;

      while i < Liste.Count do
      begin
        pData:= Liste.Items[i];
        if pData <> nil then
        begin
          case pData.typ of
          Obj_none:     begin
                        end;
          Obj_Value:    begin
                          if k = 1 then
                          begin
                            fValue:= fValue * pData.Value;
                            k:= 0;
                            for n:= i downto x-1 do
                            begin
                              if n >= 0 then
                              begin
                                pData:= Liste.Items[n];
                                if pData <> nil then dispose(pData);
                                Liste.Delete(n);
                              end;
                            end;
                            new(pNewData);
                            pNewData.Value:= fValue;
                            pNewData.c:= ' ';
                            pNewData.typ:= Obj_Value;
                            Liste.Insert(x-1,pNewData);
                            i:= -1; //restart
                          end
                          else fValue:= pData.Value;
                        end;
          Obj_Operator: begin
                          if (k = 0) and (pData.c = '*') then
                          begin
                            x:= i;
                            k:= 1;
                          end;
                        end;
          end; //case
        end;
        inc(i);
      end; //while
    end;

    //Strichrechnung
    i:= 0;
    n:= 0;
    k:= 0;
    x:= 0;
    fValue:=0.0;

    while i < Liste.Count do
    begin
      pData:= Liste.Items[i];
      if pData <> nil then
      begin
        case pData.typ of
        Obj_none:     begin
                      end;
        Obj_Value:    begin
                          if k = 1 then
                          begin
                            fValue:= fValue + pData.Value;
                            k:= 0;
                            for n:= i downto x-1 do
                            begin
                              if n >= 0 then
                              begin
                                pData:= Liste.Items[n];
                                if pData <> nil then dispose(pData);
                                Liste.Delete(n);
                              end;
                            end;
                            new(pNewData);
                            pNewData.Value:= fValue;
                            pNewData.c:= ' ';
                            pNewData.typ:= Obj_Value;
                            if x > 0 then Liste.Insert(x-1,pNewData)
                                     else Liste.Add(pNewData); //
                            i:= -1; //restart
                          end
                          else fValue:= pData.Value;
                      end;
        Obj_Operator: begin
                        if (k = 0) and (pData.c = '+') then
                        begin
                          x:= i;
                          k:= 1;
                        end;
                      end;
        end; //case
      end;
      inc(i);
    end; //while

    calc:= 0.0;
    if Liste.Count > 0 then
    begin
      pData:= Liste.Items[0];
      if pData <> nil then
      begin
        calc:= pData.Value;
        dispose(pData);
        Liste.Clear;
      end;
    end;

  end;
end;

end.
